/*
 * Programming in Java, Jacobs University, 2015
 *
 * This file demonstrates how to implement a simple chat server using
 * the non-blocking I/O framework in an event-driven style. This is
 * amazingly simple and efficient compared to the multi-threaded
 * version using the blocking I/O framework.
 *
 * This code is an edited version of sample code found online here:
 * <https://gist.github.com/Botffy/3860641>
 */

import java.net.InetSocketAddress;
import java.net.Socket;

import java.nio.ByteBuffer;
import java.nio.channels.Selector;
import java.nio.channels.SelectionKey;
import java.nio.channels.SocketChannel;
import java.nio.channels.ServerSocketChannel;

import java.io.IOException;
import java.util.Iterator;
import java.util.logging.Logger;

public class ChatServer
    implements Runnable
{
    static final int PORT = 4321;
    private static final int BUFFER_SIZE = 256;
    private static final ByteBuffer msgWelcomeBuffer
	= ByteBuffer.wrap(">> welcome to the chat room <<\n".getBytes());

    private final static Logger LOGGER
	= Logger.getLogger(ChatServer.class.getName()); 

    private static final ByteBuffer msgClientBuffer
	= ByteBuffer.wrap(">> client ".getBytes());
    private static final ByteBuffer msgJointBuffer
	= ByteBuffer.wrap(" joined <<\n".getBytes());
    private static final ByteBuffer msgLeftBuffer
	= ByteBuffer.wrap(" left <<\n".getBytes());

    private Selector selector;
    private ServerSocketChannel ssc;
    
    ChatServer(int port)
	throws IOException
    {
    	this.ssc = ServerSocketChannel.open();
    	this.ssc.socket().bind(new InetSocketAddress(port));
    	this.ssc.configureBlocking(false);
    	this.selector = Selector.open();
    	this.ssc.register(selector, SelectionKey.OP_ACCEPT);
    }
    
    public void run()
    {
	while (! Thread.currentThread().isInterrupted()) {
	    try {
		Iterator<SelectionKey> iter;
		SelectionKey key;
		while (ssc.isOpen()) {
		    selector.select();
		    iter = selector.selectedKeys().iterator();
		    while (iter.hasNext()) {
			key = iter.next();
			iter.remove();

			if (key.isValid() && key.isAcceptable()) {
			    handleAccept(key);
			}
			
			if (key.isValid() && key.isReadable()) {
			    handleRead(key);
			}
		    }
		}
	    } catch(IOException e) {
		LOGGER.warning("unexpected IO exception: " + e
			       + "; terminating");
		Thread.currentThread().interrupt();
	    }
	}
    }

    /**
     * Accepts an incoming client. Creates a new buffer used to
     * receive data from the client, sends a welcome message to the
     * client, and broadcasts the existance of the new client to all
     * clients.
     *
     * @param key the selection key for this event
     */
    
    private void handleAccept(SelectionKey key)
    	throws IOException
    {
    	SocketChannel sc = ((ServerSocketChannel) key.channel()).accept();
    	sc.configureBlocking(false);
    	sc.register(selector, SelectionKey.OP_READ,
		    ByteBuffer.allocate(BUFFER_SIZE));
	ByteBuffer client = getClientNameBuffer(sc.socket());
	ByteBuffer b = ByteBuffer.allocate(BUFFER_SIZE);

	LOGGER.info("client " + getClientName(sc.socket()) + " accepted");

	while (msgWelcomeBuffer.hasRemaining()) {
	    sc.write(msgWelcomeBuffer);
	}
    	msgWelcomeBuffer.rewind();

	b.put(msgClientBuffer).put(client).put(msgJointBuffer).flip();
	msgClientBuffer.rewind(); client.rewind(); msgJointBuffer.rewind();
	broadcast(b);
    }

    /**
     * Read data from a client. Try to process any complete lines in
     * the buffer, keep remaining data for a future read even unless
     * we run out of buffer space. In this case, the buffer is cleared
     * and the message dropped.
     *
     * @param key the selection key for this event
     */
    
    private void handleRead(SelectionKey key)
    	throws IOException
    {
    	SocketChannel sc = (SocketChannel) key.channel();
	ByteBuffer buffer = (ByteBuffer) key.attachment();
	ByteBuffer client = getClientNameBuffer(sc.socket());
	ByteBuffer b = ByteBuffer.allocate(BUFFER_SIZE + client.capacity() + 2);

	/*
	 * Get rid of clients that send excessively long lines
	 * exceeding the buffer space.
	 */
	if (buffer.remaining() == 0) {
	    LOGGER.warning("client " + getClientName(sc.socket())
			   + " exceeded input buffer; closing socket");
	    sc.close();
	    return;
	}

    	int read = sc.read(buffer);
    	if (read < 0) {
	    b.put(msgClientBuffer).put(client).put(msgLeftBuffer).flip();
	    msgClientBuffer.rewind(); client.rewind(); msgLeftBuffer.rewind();
	    broadcast(b);
	    LOGGER.info("client " + getClientName(sc.socket())
			   + " disappeared; closing socket");
	    sc.close();
	    return;
	}

	if (read > 0) {
    	    buffer.flip();
    	    byte[] bytes = new byte[buffer.limit()];
    	    buffer.get(bytes); buffer.rewind();

	    b.clear(); b.put(client).put((byte) ' '); client.rewind();
	    int offset = 0;
	    for (int i = 0; i < bytes.length; i++) {
		b.put(bytes[i]);
		if (bytes[i] == '\n') {
		    b.flip(); broadcast(b);
		    b.clear(); b.put(client).put((byte) ' '); client.rewind();
		    offset = i+1;
		}
	    }

	    if (offset != bytes.length) {
		buffer.clear();
		buffer.put(bytes, offset, bytes.length - offset);
	    }
	}
    }

    /**
     * Broadcast a message to all clients, that is all selector keys
     * with a valid socket channel.
     *
     * @param buffer contains the message to be broadcasted
     */
    
    private void broadcast(ByteBuffer buffer)
    	throws IOException
    {
    	for (SelectionKey key: selector.keys()) {
    	    if (key.isValid() && key.channel() instanceof SocketChannel) {
		SocketChannel sc = (SocketChannel) key.channel();
		try {
		    sc.write(buffer);
		    if (buffer.hasRemaining()) {
			LOGGER.warning("client " + getClientName(sc.socket())
				       + " exceeded output buffer; closing socket");
			sc.close();
		    }
		} catch (IOException e) {
		    LOGGER.warning("client " + getClientName(sc.socket())
				   + " disappeared; closing socket");
		    sc.close();
		}
    		buffer.rewind();
    	    }
    	}
    }

    /**
     * Helper to return a client identifier (name) as a ByteBuffer
     *
     * @param s socket representing the client
     */

    public static ByteBuffer getClientNameBuffer(Socket s)
    {
	String n = getClientName(s);
	return ByteBuffer.wrap(n.getBytes());
    }

    /**
     * Helper to return a client identifier (name) as a String
     *
     * @param s socket representing the client
     */

    public static String getClientName(Socket s)
    {
	return
	    "(" + s.getInetAddress().getHostName() + ":" + s.getPort() + ")";
    }

    public static void main(String[] args)
	throws IOException
    {
	ChatServer server = new ChatServer(PORT);
	(new Thread(server)).start();
    }
}
